
import fi.helsinki.cs.tmc.edutestutils.MockInOut;
import fi.helsinki.cs.tmc.edutestutils.Points;
import fi.helsinki.cs.tmc.edutestutils.ReflectionUtils;
import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Scanner;
import org.junit.Assert;
import static org.junit.Assert.*;
import org.junit.Test;

public class FirstPackagesTest {

    @Test
    @Points("10-08.1")
    public void uiInterfaceExists() {
        Class clazz = null;
        try {
            clazz = ReflectionUtils.findClass("mooc.ui.userInterface");
        } catch (Throwable e) {
            fail("Create the package mooc.ui and add a class called userInterface in it:\n\n"
                    + "public interface userInterface{\n  \\\\...\n}");
        }

        if (clazz == null) {
            fail("Please ensure that you have created an interface to the package mooc.ui and that it is defined as public.");
        }

        if (!clazz.isInterface()) {
            fail("please ensure that userInterface is defined as an interface");
        }

        boolean found = false;
        for (Method m : clazz.getMethods()) {
            if (!m.getReturnType().equals(void.class)) {
                continue;
            }

            if (!m.getName().equals("update")) {
                continue;
            }

            found = true;
        }

        if (!found) {
            fail("Please ensure that the interface userInterface has the method: void update()");
        }
    }

    /*
     *
     */
    @Test
    @Points("10-08.2")
    public void textInterfaceExists() {
        Class clazz = ReflectionUtils.findClass("mooc.ui.textInterface");

        if (clazz == null) {
            fail("Please ensure that you have created the class textInterface in the package mooc.ui and that it is defined as public.");
        }

        Class kali = ReflectionUtils.findClass("mooc.ui.userInterface");
        for (Class iface : clazz.getInterfaces()) {
            if (iface.equals(kali)) {
                return;
            }
        }

        fail("Please ensure that the class textInterface implements the interface userInterface");
    }

    @Test
    @Points("10-08.2")
    public void textInterfacePrints() {
        Class clazz = ReflectionUtils.findClass("mooc.ui.textInterface");
        Object kali = createTextInterface();

        MockInOut io = new MockInOut("");

        Method m = ReflectionUtils.requireMethod(clazz, "update");
        try {
            m.invoke(kali);
        } catch (Throwable t) {
            fail("There was an error executing your code. Test what happens when you execute it yourself.\n"
                    + "userInterface kali = new textInterface();\n"
                    + "kali.update()\n");
        }

        Assert.assertTrue("Please ensure that the update method of the textInterface class prints something. \n"
                + "What it printed out now:\n" + io.getOutput(), io.getOutput() != null && io.getOutput().length() > 1);
        Assert.assertTrue("The update-method of the textInterface-class should print a newline \n"
                + "What it printed out now:\n" + io.getOutput(), io.getOutput() != null && io.getOutput().contains("\n"));
        Assert.assertTrue("The update-method of the textInterface-class should only print one line. \n"
                + "NWhat it printed out now:\n" + io.getOutput(), io.getOutput() != null && io.getOutput().split("\n").length == 1);
    }

    /*
     *
     */
    @Test
    @Points("10-08.3")
    public void applicationLogicExists() {
        Class applicationLogicClass = ReflectionUtils.findClass("mooc.logic.applicationLogic");

        if (applicationLogicClass == null) {
            fail("Please ensure that you have created the class 'applicationLogic' in the package mooc.logic and that it is defined as public.");
        }
        Class kaliClass = ReflectionUtils.findClass("mooc.ui.userInterface");

        Constructor constructor = ReflectionUtils.requireConstructor(applicationLogicClass, kaliClass);
        if (constructor == null) {
            fail("Please ensure that the class applicationLogic has the constructor 'public applicationLogic(userInterface ui)'");
        }

        Method m = null;

        try {
            m = ReflectionUtils.requireMethod(applicationLogicClass, "execute", int.class);
        } catch (Throwable t) {
            fail("The applicationLogic-class is missing the method: public void execute(int times)");
        }

        assertTrue("The applicationLogic-class is missing the method: public void execute(int times)", m.toString().contains("public"));
    }

    @Test
    @Points("10-08.3")
    public void limitVariableCount() {
        sanitezationCheck("mooc.logic.applicationLogic", 1, "a userInterface-type object variable");
    }

    @Test
    @Points("10-08.3")
    public void creatingApplicationLogicObject() throws IllegalArgumentException, IllegalAccessException {
        Class applicationLogicClass = ReflectionUtils.findClass("mooc.logic.applicationLogic");
        Object applicationLogic = createApplicationLogicObject();
        if (applicationLogic == null) {
            fail("Onhan luokalla Sovelluslogiikka konstruktori public Sovelluslogiikka(Kayttoliittyma kayttoliittyma)");
            fail("Please ensure that the applicationLogic-class has the constructor 'public applicationLogic(userInterface ui)'");
        }

        Method executeMethod = ReflectionUtils.requireMethod(applicationLogicClass, "execute", int.class);

        Class clazz = ReflectionUtils.findClass("mooc.ui.textInterface");
        Object kali = createTextInterface();

        MockInOut io = new MockInOut("");

        Method m = ReflectionUtils.requireMethod(clazz, "update");
        try {
            m.invoke(kali);
        } catch (Throwable t) {
            fail("There was an error executing your code. Test what happens when you execute it yourself.\n"
                    + "userInterface kali = new textInterface();\n"
                    + "kali.update()\n");
        }

        Assert.assertTrue("Please ensure that the update method of the textInterface class prints something.", io.getOutput() != null && io.getOutput().length() > 5);
        String output = io.getOutput();

        Field[] fields = ReflectionUtils.findClass("mooc.logic.applicationLogic").getDeclaredFields();
        assertTrue("Please ensure that the applicationLogic-class has a userInterface-type objectvariable.", fields.length == 1);
        assertTrue("Please ensure that the applicationLogic-class has a userInterface-type objectvariable.", fields[0].toString().contains("userInterface"));
        fields[0].setAccessible(true);
        assertFalse("the objectvariable of applicationLogic " + fields[0].toString().replace("mooc.logic.applicationLogic", "") + " "
                + "null. \nPlease set it to the variable passed from the appLogic's constructor!", fields[0].get(applicationLogic) == null);

        try {
            executeMethod.invoke(applicationLogic, 3);
        } catch (Throwable t) {
            fail("There was an error executing your code. Test what happens when you execute it yourself.\n"
                    + "userInterface kali = new textInterface();\n"
                    + "applicationLogic app = new applicationLogic(kali);\n"
                    + "app.execute(3);");
        }

        String executionOutput = io.getOutput().substring(output.length());

        assertTrue("With the code\n"
                + "userInterface kali = new textInterface();\n"
                + "applicationLogic app = new applicationLogic(kali);\n"
                + "app.execute(3);\n"
                + "6 lines should be printed.\nWhat was printed now: \n" + executionOutput, executionOutput.split("\n").length > 5 && executionOutput.split("\n").length < 8);

        if (executionOutput.length() < output.length() * 3) {
            fail("Please ensure that you call the userInterface-interfaces update-method exactly as many times as given by the variable 'times' in the applicationLogic's execute-methods loop construct.");
        }
    }

    @Test
    @Points("10-08.3")
    public void anotherApplication() throws IllegalArgumentException, IllegalAccessException {
        Class appLogicClass = ReflectionUtils.findClass("mooc.logic.applicationLogic");
        Object appLogic = createApplicationLogicObject();
        if (appLogic == null) {
            fail("Please ensure that the applicationLogic-class has the constructor 'public applicationLogic(userInterface ui)'");
        }

        Method executeMethod = ReflectionUtils.requireMethod(appLogicClass, "execute", int.class);

        Class clazz = ReflectionUtils.findClass("mooc.ui.textInterface");
        Object kali = createTextInterface();

        MockInOut io = new MockInOut("");

        Method m = ReflectionUtils.requireMethod(clazz, "update");
        try {
            m.invoke(kali);
        } catch (Throwable t) {
            fail("There was an error executing your code. Test what happens when you execute it yourself.\n"
                    + "userInterface kali = new textInterface();\n"
                    + "kali.update()\n");
        }

        Assert.assertTrue("Please ensure that the update-method of the textInterface-class print something", io.getOutput() != null && io.getOutput().length() > 5);
        String output = io.getOutput();

        Field[] kentat = ReflectionUtils.findClass("mooc.logic.applicationLogic").getDeclaredFields();
        assertTrue("Please ensure that the applicationLogic-class has a userInterface-type objectvariable", kentat.length == 1);
        assertTrue("Please ensure that the applicationLogic-class has a userInterface-type objectvariable", kentat[0].toString().contains("ui"));
        kentat[0].setAccessible(true);
        assertFalse("the applicationLogic's variable " + kentat[0].toString().replace("mooc.logiikka.Sovelluslogiikka.", "") + " "
                + "null. \nPlease set it to the variable passed from the appLogic's constructor!", kentat[0].get(appLogic) == null);

        try {
            executeMethod.invoke(appLogic, 5);
        } catch (Throwable t) {
            fail("There was an error executing your code. Test what happens when you execute it yourself.\n"
                    + "userInterface kali = new textInterface();\n"
                    + "applicationLogic app = new applicationLogic(kali);\n"
                    + "app.execute(3);");
        }

        String executionOutput = io.getOutput().substring(output.length());

        assertTrue("With the code:\n"
                + "userInterface kali = new textInterface();\n"
                + "applicationLogic app = new applicationLogic(kali);\n"
                + "app.execute(5);\n"
                + "6 lines should be printed.\nWhat was printed now: \n" + executionOutput, executionOutput.split("\n").length > 9 && executionOutput.split("\n").length < 12);
    }

    @Test
    @Points("10-08.3")
    public void applicationCallsTheUI() {
        Field[] kentat = ReflectionUtils.findClass("mooc.logic.applicationLogic").getDeclaredFields();
        String muuttuja = kentat[0].toString();
        muuttuja = muuttuja.substring(muuttuja.lastIndexOf(".") + 1);
        assertTrue("Luokan Sovelluslogiikka metodin suorita on kutsuttava käyttöliittymän metodia paivita!", includesCall(muuttuja + ".update()"));
        assertTrue("The execute-method of the applicationLogic-class must call for the update-method of the userInterface", includesCall(muuttuja + ".update()"));
    }

    private boolean includesCall(String call) {

        try {
            Scanner scanner = new Scanner(new File("src/main/java/mooc/logic/applicationLogic.java"));
            int inMethod = 0;

            while (scanner.hasNextLine()) {

                String line = scanner.nextLine();

                if (line.indexOf("//") > -1) {
                    continue;
                }

                if (inMethod > 0 && line.contains(call)) {
                    return true;
                }

                if (line.contains("void") && line.contains("execute")) {
                    inMethod++;
                } else if (inMethod > 0) {
                    if (line.contains("{") && !line.contains("}")) {
                        inMethod++;
                    }

                    if (line.contains("}") && !line.contains("{")) {
                        inMethod--;
                    }
                }

            }

        } catch (Exception e) {
            fail(e.getMessage());
        }

        return false;
    }

    private Object createTextInterface() {
        Class clazz = ReflectionUtils.findClass("mooc.ui.textInterface");
        Object kali = null;
        try {
            kali = ReflectionUtils.invokeConstructor(ReflectionUtils.requireConstructor(clazz));
        } catch (Throwable ex) {
        }

        if (kali == null) {
            fail("Please ensure that the class textInterface has a constructor with 0 parameters");
        }

        return kali;
    }

    private Object createApplicationLogicObject() {
        Class applicationLogicClass = ReflectionUtils.findClass("mooc.logic.applicationLogic");

        if (applicationLogicClass == null) {
            fail("Please ensure that you have created the class applicationLogic, and that it is defined as public.");
        }
        Class kaliClass = ReflectionUtils.findClass("mooc.ui.userInterface");

        Constructor constructor = ReflectionUtils.requireConstructor(applicationLogicClass, kaliClass);

        try {
            return ReflectionUtils.invokeConstructor(constructor, createTextInterface());
        } catch (Throwable ex) {
        }

        return null;
    }

    private void sanitezationCheck(String klassName, int n, String m) throws SecurityException {
        Field[] fields = ReflectionUtils.findClass(klassName).getDeclaredFields();

        for (Field CurrentField : fields) {
            assertFalse("you dont need \"static variables\", delete the variable " + field(CurrentField.toString(), klassName) + " from the class " + klassName , CurrentField.toString().contains("static") && !CurrentField.toString().contains("final"));
            //assertTrue("luokan kaikkien oliomuuttujien näkyvyyden tulee olla private, muuta luokalta " + klassName + " löytyi: " + field(CurrentField.toString(), klassName), CurrentField.toString().contains("private"));
            assertTrue("all variables of the classes should be private, but the class " + klassName + " had: " + field(CurrentField.toString(), klassName) + " as a public variable.", CurrentField.toString().contains("private"));
        }

        if (fields.length > 1) {
            int var = 0;
            for (Field field : fields) {
                if (!field.toString().contains("final")) {
                    var++;
                }
            }
            //assertTrue("et tarvitse " + klassName + "-luokalle kuin " + m + ", poista ylimääräiset", var <= n);
            assertTrue("You sholdn't need more than " + m + " for the class " + klassName + ", please delete the unnecessary ones", var <= n);
        }
    }

    private String field(String toString, String klassName) {
        return toString.replace(klassName + ".", "").replace("java.lang.", "").replace("java.util.", "");
    }
}
